########################################################################
#
# tk kanji configuration
#
# Copyright 1999 by Roger E. Critchlow Jr, Santa Fe, New Mexico, USA,
# rec@elf.org -- http://www.elf.org
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA
#
########################################################################

########################################################################
#
# kanjidic parsing -
#  initially just split at new line
#  and index by [lindex $line 0]
#  which will give a unicode index
#  Then split lines at space and classify
#  results by the first character.
#  actually, second word is hexadecimal JIS,
#  either jisx-0208 (3021 .. 7426)
#  or jisx-0212 (3021 .. 6d63)
#  depending on the file read,
#
#  then classify by leading character
# 
# each record has the identifying character(s), a name,
# a category and an explanatory name.
# The categories are
#	class - maps kanji to equivalence classes
#	index - maps kanji, or a subset, to unique identifiers
#	trans - translates kanji to english, pinyin, ...
#	refer - references another kanji record
#
array set kanjifields {
    B     { class {Radical} {Bushu radical number} }
    C     { class {Radical(2)} {Historical or classical radical number} }
    F     { index {Frequency} {Frequency of use ranking} }
    G     { class {Jouyou} {Jouyou grade-level} }
    H     { index {Halpern index} {Halpern's New Japanese-English Character Dictionary index number} }
    N     { index {Nelson index} {Nelson's Modern Reader's Japanese-English Character Dictionary index number} }
    DR    { index {De Roo code} {Codes developed by De Roo} }
    E     { index {Henshall index} {Henshall, A Guide To Remembering Japanese Characters index number} }
    IN    { index {S&H index(2)} {Spahn and Hadamitzky, Kanji and Kana index number} }
    I     { class {S&H index} {Spahn and Hadamitzky, The Kanji Dictionary index code} }
    J0    { index {JIS0} {JISX-0208 code point} }
    J1    { index {JIS1} {JISX-0212 code point} }
    K     { index {GKD index} {Gakken Kanji Dictionary index number} }
    L     { index {Heisig index} {Heisig, Remembering The Kanji index number} }
    MN    { index {Morohashi index} {Morohashi index number} }
    MP    { index {Morohashi page} {Morohashi volume.page numbers} }
    O     { index {ONeill index} {O'Neill, Japanese Names index number} }
    P     { class {SKIP} {SKIP pattern code} }
    Q     { class {Four Corner} {Four Corner code} }
    S     { class {Strokes} {Stroke count} }
    U     { index {Unicode} {Unicode code point} }
    V     { index {Nelson index(2)} {New Nelson Japanese-English Character Dictionary index number} }
    W     { trans {Korean} {Romanized Korean reading(s)} }
    Y     { trans {Pinyin} {Pinyin reading(s)} }
    X     { refer {Crossref} {Cross reference code} }
    Z     { refer {Misclass} {Mis-classification code} }
}
#
# these are constructed fields, they are untagged or shift tagged
# in the input.  When we split the record into property lists, we
# insert these tags to identify the untagged fields.
#
array set kanjifields {
    ON    { trans {On} {On readings in katakana} }
    ONT1  { trans {On name} {On name forms in katakana} }
    ONT2  { trans {On radical} {On radical name if not earlier} }
    KUN   { trans {Kun} {Kun readings in hiragana} }
    KUNT1 { trans {Kun name} {Kun name forms in katakana} }
    KUNT2 { trans {Kun radical} {Kun radical name if not earlier} }
    ENG   { trans {English} {English translation} }
}
#
# get the field type
#
proc tkj-get-field-type {f} {
    global kanjifields
    return [lindex $kanjifields($f) 0]
}
#
# get the field label
#
proc tkj-get-field-label {f} {
    global kanjifields
    return [lindex $kanjifields($f) 1]
}
#
# get the field description
#
proc tkj-get-field-description {f} {
    global kanjifields
    return [lindex $kanjifields($f) 2]
}
#
# translate some field values into more useful strings
#
array set kanjifieldvalues {
    G  {All Jouyou}
    G1 Jouyou-1
    G2 Jouyou-2
    G3 Jouyou-3
    G4 Jouyou-4
    G5 Jouyou-5
    G6 Jouyou-6
    G8 Jouyou
    G9 Jinmeiyou
}
#
# return a useful string for a field value
#    
proc tkj-get-field-value-label {f v} {
    global kanjifieldvalues
    if {[info exists kanjifieldvalues($f$v)]} {
	return $kanjifieldvalues($f$v)
    }
    return $v
}
#
# load a kanji dictionary
#   index each record under the unicode for the kanji
#   scan the records to
#     1) accumulate the members of each index set
#     2) accumulate the values and members of each class set
#     3) rewrite the JISX values to identify encoding
#
proc tkj-load-dictionary {file jiskey} {
    global kanji
    global kanjiclass

    set n 1
    tkj-status "loading [file tail $file]"

    set fp [open $file]
    fconfigure $fp -encoding euc-jp
    gets $fp line;			# skip header line

    while {[gets $fp line] >= 0} {
	# post status
	if {[incr n] % 100 == 0} {
	    tkj-status "loading [file tail $file] $n lines read"
	}
	# use unicode to define key, because Tcl won't decode EUC-3
	set k [subst \\u[string range [lindex $line 2] 1 end]]
	# minimal rebuild of input record
	set kanji($k) [concat [list $jiskey[lindex $line 1]] [lrange $line 2 end]];
	# remember the input list
	lappend kanjiclass(all) $k
	# build up class and index membership lists
	foreach i $kanji($k) {
	    switch -glob -- $i {
		B*  { lappend kanjiclass(B)  $k; lappend kanjiclass($i) $k; continue }
		C*  { lappend kanjiclass(C)  $k; lappend kanjiclass($i) $k; continue }
		DR* { lappend kanjiclass(DR) $k; continue }
		E*  { lappend kanjiclass(E)  $k; continue }
		F*  { lappend kanjiclass(F)  $k; continue }
		G*  { lappend kanjiclass(G)  $k; lappend kanjiclass($i) $k; continue }
		H*  { lappend kanjiclass(H)  $k; continue }
		IN* { lappend kanjiclass(IN) $k; continue }
		I*  { lappend kanjiclass(I)  $k; lappend kanjiclass([lindex [split $i .] 0]) $k; continue }
		J0* { lappend kanjiclass(J0) $k; continue }
		J1* { lappend kanjiclass(J1) $k; continue }
		K*  { lappend kanjiclass(K)  $k; continue }
		L*  { lappend kanjiclass(L)  $k; continue }
		MN* { lappend kanjiclass(MN) $k; continue }
		MP* { lappend kanjiclass(MP) $k; continue }
		N*  { lappend kanjiclass(N)  $k; continue }
		O*  { lappend kanjiclass(O)  $k; continue }
		P*  { lappend kanjiclass(P)  $k; lappend kanjiclass($i) $k; continue }
		Q*  { lappend kanjiclass(Q)  $k; lappend kanjiclass([lindex [split $i .] 0]) $k; continue }
		S*  { lappend kanjiclass(S)  $k; lappend kanjiclass($i) $k; continue }
		U*  { lappend kanjiclass(U)  $k; continue }
		V*  { lappend kanjiclass(V)  $k; continue }
		W*  { lappend kanjiclass(W)  $k; continue }
		Y*  { lappend kanjiclass(Y)  $k; continue }
		X*  { lappend kanjiclass(X)  $k; continue }
		Z*  { lappend kanjiclass(Z)  $k; continue }
		*   { break }
	    }
	    break
	}
    }
    close $fp
    tkj-status ""
}
#
# get the records of a kanji split into a property list of
# names and values
#
proc get-record {k} {
    global kanji
    global kanjisplit
    if { ! [info exists kanjisplit($k)]} {
	set info 1
	set spec {}
	set vals {}
	foreach i $kanji($k) {
	    if {$info} { 
		switch -glob -- $i {
		    DR* -
		    IN* -
		    J0* -
		    J1* -
		    MN* -
		    MP* {
			lappend vals [string range $i 0 1] [string range $i 2 end]
			continue
		    }
		    B* -
		    C* -
		    E* -
		    F* -
		    G* -
		    H* -
		    I* -
		    K* -
		    L* -
		    N* -
		    O* -
		    P* -
		    Q* -
		    S* -
		    U* -
		    V* -
		    W* -
		    Y* -
		    X* -
		    Z*  {
			lappend vals [string range $i 0 0] [string range $i 1 end]
			continue
		    }
		}
		set info 0
	    }
	    switch -regexp -- $i {
		{^-?[\u3040-\u309f]+(\.[\u3040-\u309f]+)?-?$} {
		    # Hirigana
		    lappend vals KUN$spec $i
		    continue
		}
		{^-?[\u30a0-\u30ff]+(\.[\u30a0-\u30ff])?-?$} {
		    # Katakana
		    lappend vals ON$spec $i
		    continue
		}
		^T1$ -
		^T2$ {
		    set spec $i
		    continue
		}
		default {
		    # English
		    lappend vals ENG $i
		}
	    }
	}
	set kanji($k) $vals
	set kanjisplit($k) {}
    }
    return $kanji($k)
}
#
# get a field(s) of a record matching a field key
# return a list of results
#
proc tkj-get-field {k f} {
    set vals {}
    foreach {name value} [get-record $k] {
	if {[string compare $name $f] == 0} {
	    lappend vals $value
	}
    }
    return $vals
}
#
# get the set indexed by key $f
# sorted by the index value
#
proc tkj-get-index {f} {
    global kanjiclass
    global kanjiclasssort
    if { ! [info exists kanjiclasssort($f)] && [string compare $f all] != 0} {
	foreach k $kanjiclass($f) {
	    lappend invert([lindex [tkj-get-field $k $f] 0]) $k
	}
	set kanjiclass($f) {}
	foreach v [lsort -integer [array names invert]] {
	    foreach k $invert($v) {
		lappend kanjiclass($f) $k
	    }
	}
	set kanjiclasssort($f) {}
    }
    return $kanjiclass($f)
}
#
# get the class members of a class
#
proc tkj-get-class {c} {
    global kanjiclass
    set kanjiclass($c)
}
#
# get the number of members of a class
#
proc tkj-class-size {c} {
    global kanjiclass
    llength $kanjiclass($c)
}
#
# get the names of classes matching a pattern
#
proc tkj-class-names {p} {
    global kanjiclass
    array names kanjiclass $p
}
#
#
#
proc tkj-class-exists {c} {
    global kanjiclass
    info exists kanjiclass($c)
}

